package org.zeroturnaround.javarebel.integration.spring;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.springframework.aop.framework.Advised;
import org.springframework.beans.factory.config.ConfigurableListableBeanFactory;
import org.springframework.context.annotation.ComponentScanBeanDefinitionParser;
import org.zeroturnaround.javarebel.ClassEventListener;
import org.zeroturnaround.javarebel.Logger;
import org.zeroturnaround.javarebel.LoggerFactory;

/**
 * This is the central class of the JavaRebel Spring plugin. 
 * It is the entry point for both bean definition reloading ({@link #onClassEvent(int, Class)}) 
 * and bean reconfiguration ({@link #reloadBeanDefinitions()}.  
 * 
 * @author Jevgeni Kabanov
 */
public final class SpringBeanReloader implements ClassEventListener {
  
  private static final Logger log = LoggerFactory.getInstance();
  
  private static final boolean isAnnotationBeanDefinitionEnabled = isAnnotationBeanDefinitionEnabled();
  
  private static ThreadLocal changedBeans = new ThreadLocal();

  private ConfigurableListableBeanFactory bf;    

  public SpringBeanReloader(ConfigurableListableBeanFactory bf) {
    this.bf = bf;
  }

  /**
   * This method will reconfigure the singletons when their class changes. 
   * It will delay reconfiguration if we are running in a {@link #reloadBeanDefinitions()}
   * method to be sure we are reconfiguring against the latest definitions.
   */
  public void onClassEvent(int eventType, Class klass) {
    if (changedBeans.get() == null)
      reloadBeanDefinitions();

    String[] singletonNames = bf.getSingletonNames();

    for (int i = 0; i < singletonNames.length; i++) {
      final String singletonName = singletonNames[i];
      Object bean = bf.getBean(singletonName);
      
      // Since Spring 2.0 FactoryBean#getObject() may also return null
      if (bean != null) {
        if (bean instanceof Advised) {
          try {
            bean = ((Advised) bean).getTargetSource().getTarget();
          } catch (Exception e) {
            LoggerFactory.getInstance().error(e);
          }
        }
        
        final Object theBean = bean;
        
        if (klass.isAssignableFrom(bean.getClass())) {
          if (changedBeans.get() == null) {
            // Do not delay it
            log.echo("JavaRebel-Spring: Reconfiguring bean '" + singletonName + "' [" + theBean.getClass().getName() + "]");
            bf.configureBean(theBean, singletonName);
          } else {
            // Delay the action
            Runnable run = new Runnable() {
              public void run() {
                log.echo("JavaRebel-Spring: Reconfiguring bean '" + singletonName + "' [" + theBean.getClass().getName() + "]");
                bf.configureBean(theBean, singletonName);
              }
            };
            ((List) changedBeans.get()).add(run);
          }
        }
      }
    }
  }

  /**
   * This method will reload the bean definitions and apply postprocessing. 
   * As it touches classes it may cause them to reload if they were changed.
   */
  public static void reloadBeanDefinitions() {
    changedBeans.set(new ArrayList());
    try {
      XmlBeanDefinitionManager.checkAndReload();
      
      if (isAnnotationBeanDefinitionEnabled) {
        AnnotationBeanDefinitionManager.rescan();
      }
    }
    finally {
      List toRun = (List) changedBeans.get();
      changedBeans.set(null);
      
      for (Iterator i = toRun.iterator(); i.hasNext();) {
        Runnable run = (Runnable) i.next();
        run.run();
      }                 
    }
  }
  
  /**
   * @return <code>true</code> if AnnotationBeanDefinitionManager can be used.
   */
  private static boolean isAnnotationBeanDefinitionEnabled() {
    try {
      ComponentScanBeanDefinitionParser.class.getName();
      return true;
    } catch (NoClassDefFoundError e) {
      return false;
    }
  }
}